import asyncio
import logging
import signal
import sys
import traceback
import typing

import HABApp
from HABApp.__cmd_args__ import parse_args


def register_signal_handler():
    def shutdown_handler(sig, frame):
        print('Shutting down ...')
        HABApp.runtime.shutdown.request_shutdown()

    # register shutdown helper
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)


def main() -> typing.Union[int, str]:

    # This has do be done before we create HABApp because of the possible sleep time
    cfg_folder = parse_args()

    log = logging.getLogger('HABApp')

    try:
        app = HABApp.runtime.Runtime()
        register_signal_handler()

        # start workers
        try:
            asyncio.ensure_future(app.start(cfg_folder))
            HABApp.core.const.loop.run_forever()
        except asyncio.CancelledError:
            pass

    except HABApp.config.InvalidConfigException:
        pass
    except Exception as e:
        for line in traceback.format_exc().splitlines():
            log.error(line)
            print(e)
        return str(e)
    finally:
        # Sleep to allow underlying connections of aiohttp to close
        # https://aiohttp.readthedocs.io/en/stable/client_advanced.html#graceful-shutdown
        HABApp.core.const.loop.run_until_complete(asyncio.sleep(1))
        HABApp.core.const.loop.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())



import asyncio
from pathlib import Path

import HABApp.config
import HABApp.core
import HABApp.mqtt.mqtt_connection
import HABApp.parameters.parameter_files
import HABApp.rule_manager
import HABApp.util
import eascheduler
from HABApp.core.wrapper import process_exception
from HABApp.openhab import connection_logic as openhab_connection
from HABApp.runtime import shutdown
from HABApp.core.context import async_context

import HABApp.rule.interfaces._http


class Runtime:

    def __init__(self):
        self.config: HABApp.config.Config = None

        # Rule engine
        self.rule_manager: HABApp.rule_manager.RuleManager = None

        # Async Workers & shutdown callback
        shutdown.register_func(HABApp.core.WrappedFunction._WORKERS.shutdown, msg='Stopping workers')

    async def start(self, config_folder: Path):
        try:
            token = async_context.set('HABApp startup')

            # setup exception handler for the scheduler
            eascheduler.set_exception_handler(lambda x: process_exception('HABApp.scheduler', x))

            # Start Folder watcher!
            HABApp.core.files.watcher.start()

            self.config_loader = HABApp.config.HABAppConfigLoader(config_folder)

            await HABApp.core.files.setup()

            # generic HTTP
            await HABApp.rule.interfaces._http.create_client()

            # openhab
            openhab_connection.setup()

            # Parameter Files
            await HABApp.parameters.parameter_files.setup_param_files()

            # Rule engine
            self.rule_manager = HABApp.rule_manager.RuleManager(self)
            await self.rule_manager.setup()

            # MQTT
            HABApp.mqtt.mqtt_connection.setup()
            HABApp.mqtt.mqtt_connection.connect()

            await openhab_connection.start()

            shutdown.register_func(HABApp.core.const.loop.stop, msg='Stopping asyncio loop')

            async_context.reset(token)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            process_exception('Runtime.start', e)
            await asyncio.sleep(1)  # Sleep so we can do a graceful shutdown
            shutdown.request_shutdown()



import asyncio
import datetime
import logging
import random
import re
import sys
import traceback
import typing
import warnings
import weakref
from typing import Iterable, Union

import HABApp
import HABApp.core
import HABApp.openhab
import HABApp.rule_manager
import HABApp.util
from HABApp.core.events import AllEvents
from HABApp.core.items.base_item import BaseItem, TYPE_ITEM, TYPE_ITEM_CLS
from HABApp.rule import interfaces
from HABApp.rule.scheduler import HABAppSchedulerView as _HABAppSchedulerView
from .interfaces import async_subprocess_exec

log = logging.getLogger('HABApp.Rule')


# Func to log deprecation warnings
def send_warnings_to_log(message, category, filename, lineno, file=None, line=None):
    log.warning('%s:%s: %s:%s' % (filename, lineno, category.__name__, message))
    return


# Setup deprecation warnings
warnings.simplefilter('default')
warnings.showwarning = send_warnings_to_log


class Rule:
    def __init__(self):

        # get the variables from the caller
        depth = 1
        while True:
            try:
                __vars = sys._getframe(depth).f_globals
            except ValueError:
                raise RuntimeError('Rule files are not meant to be executed directly! '
                                   'Put the file in the HABApp "rule" folder and HABApp will load it automatically.')

            depth += 1
            if '__HABAPP__RUNTIME__' in __vars:
                __runtime__ = __vars['__HABAPP__RUNTIME__']
                __rule_file__ = __vars['__HABAPP__RULE_FILE__']
                break

        # variable vor unittests
        test = __vars.get('__UNITTEST__', False)

        # this is a list which contains all rules of this file
        __vars['__HABAPP__RULES'].append(self)

        assert isinstance(__runtime__, HABApp.runtime.Runtime)
        self.__runtime: HABApp.runtime.Runtime = __runtime__

        if not test:
            assert isinstance(__rule_file__, HABApp.rule_manager.RuleFile)
        self.__rule_file: HABApp.rule_manager.RuleFile = __rule_file__

        self.__event_listener: typing.List[HABApp.core.EventBusListener] = []
        self.__unload_functions: typing.List[typing.Callable[[], None]] = []
        self.__cancel_objs: weakref.WeakSet = weakref.WeakSet()

        # schedule cleanup of this rule
        self.register_on_unload(self.__cleanup_rule)
        self.register_on_unload(self.__cleanup_objs)

        # scheduler
        self.run: _HABAppSchedulerView = _HABAppSchedulerView(self)

        # suggest a rule name if it is not
        self.rule_name: str = self.__rule_file.suggest_rule_name(self)

        # interfaces
        self.async_http = interfaces.http
        self.mqtt: HABApp.mqtt.interface = HABApp.mqtt.interface
        self.oh: HABApp.openhab.interface = HABApp.openhab.interface
        self.openhab: HABApp.openhab.interface = self.oh

    @HABApp.core.wrapper.log_exception
    def __cleanup_objs(self):
        while self.__cancel_objs:
            # we log each error as warning
            with HABApp.core.wrapper.ExceptionToHABApp(log, logging.WARNING):
                obj = self.__cancel_objs.pop()
                obj.cancel()

    @HABApp.core.wrapper.log_exception
    def __cleanup_rule(self):
        # Important: set the dicts to None so we don't schedule a future event during _cleanup.
        # If dict is set to None we will crash instead but it is no problem because everything gets unloaded anyhow
        event_listeners = self.__event_listener
        self.__event_listener = None

        # Actually remove the listeners/events
        for listener in event_listeners:
            HABApp.core.EventBus.remove_listener(listener)

        # Unload the scheduler
        self.run._scheduler.cancel_all()
        return None

    def post_event(self, name, event):
        """
        Post an event to the event bus

        :param name: name or item to post event to
        :param event: Event class to be used (must be class instance)
        :return:
        """
        assert isinstance(name, (str, HABApp.core.items.BaseValueItem)), type(name)
        return HABApp.core.EventBus.post_event(
            name.name if isinstance(name, HABApp.core.items.BaseValueItem) else name,
            event
        )

    def listen_event(self, name: Union[HABApp.core.items.BaseValueItem, str],
                     callback: typing.Callable[[typing.Any], typing.Any],
                     event_type: Union[typing.Type['HABApp.core.events.AllEvents'],
                                       'HABApp.core.events.EventFilter', typing.Any] = AllEvents
                     ) -> HABApp.core.EventBusListener:
        """
        Register an event listener

        :param name: item or name to listen to. Use None to listen to all events
        :param callback: callback that accepts one parameter which will contain the event
        :param event_type: Event filter. This is typically :class:`~HABApp.core.events.ValueUpdateEvent` or
            :class:`~HABApp.core.events.ValueChangeEvent` which will also trigger on changes/update from openhab
            or mqtt. Additionally it can be an instance of :class:`~HABApp.core.events.EventFilter` which additionally
            filters on the values of the event. There are also templates for the most common filters, e.g.
            :class:`~HABApp.core.events.ValueUpdateEventFilter` and :class:`~HABApp.core.events.ValueChangeEventFilter`
        """
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))
        name = name.name if isinstance(name, HABApp.core.items.BaseValueItem) else name

        if isinstance(event_type, HABApp.core.events.EventFilter):
            listener = event_type.create_event_listener(name, cb)
        else:
            listener = HABApp.core.EventBusListener(name, cb, event_type)

        self.__event_listener.append(listener)
        HABApp.core.EventBus.add_listener(listener)
        return listener

    def execute_subprocess(self, callback, program, *args, capture_output=True):
        """Run another program

        :param callback: |param_scheduled_cb| after process has finished. First parameter will
                         be an instance of :class:`~HABApp.rule.FinishedProcessInfo`
        :param program: program or path to program to run
        :param args: |param_scheduled_cb_args|
        :param capture_output: Capture program output, set to `False` to only capture return code
        :return:
        """

        assert isinstance(program, str), type(program)
        cb = HABApp.core.WrappedFunction(callback, name=self._get_cb_name(callback))

        asyncio.run_coroutine_threadsafe(
            async_subprocess_exec(cb.run, program, *args, capture_output=capture_output),
            HABApp.core.const.loop
        )

    def get_rule(self, rule_name: str) -> 'Union[Rule, typing.List[Rule]]':
        assert rule_name is None or isinstance(rule_name, str), type(rule_name)
        return self.__runtime.rule_manager.get_rule(rule_name)

    def register_on_unload(self, func: typing.Callable[[], typing.Any]):
        """Register a function with no parameters which will be called when the rule is unloaded.
        Use this for custom cleanup functions.

        :param func: function which will be called
        """
        assert callable(func)
        assert func not in self.__unload_functions, 'Function was already registered!'
        self.__unload_functions.append(func)

    def register_cancel_obj(self, obj):
        """Add a ``weakref`` to an obj which has a ``cancel`` function.
        When the rule gets unloaded the cancel function will be called (if the obj was not already garbage collected)

        :param obj:
        """
        self.__cancel_objs.add(obj)

    @staticmethod
    def get_items(type: Union[typing.Tuple[TYPE_ITEM_CLS, ...], TYPE_ITEM_CLS] = None,
                  name: Union[str, typing.Pattern[str]] = None,
                  tags: Union[str, Iterable[str]] = None,
                  groups: Union[str, Iterable[str]] = None,
                  metadata: Union[str, typing.Pattern[str]] = None,
                  metadata_value: Union[str, typing.Pattern[str]] = None,
                  ) -> Union[typing.List[TYPE_ITEM], typing.List[BaseItem]]:
        """Search the HABApp item registry and return the found items.

        :param type: item has to be an instance of this class
        :param name: str (will be compiled) or regex that is used to search the Name
        :param tags: item must have these tags (will return only instances of OpenhabItem)
        :param groups: item must be a member of these groups (will return only instances of OpenhabItem)
        :param metadata: str (will be compiled) or regex that is used to search the metadata (e.g. 'homekit')
        :param metadata_value: str (will be compiled) or regex that is used to search the metadata value
                               (e.g. 'TargetTemperature')
        :return: Items that match all the passed criteria
        """

        if name is not None and isinstance(name, str):
            name = re.compile(name, re.IGNORECASE)
        if metadata is not None and isinstance(metadata, str):
            metadata = re.compile(metadata, re.IGNORECASE)
        if metadata_value is not None and isinstance(metadata_value, str):
            metadata_value = re.compile(metadata_value, re.IGNORECASE)

        _tags, _groups = None, None
        if tags is not None:
            _tags = set(tags) if not isinstance(tags, str) else {tags}
        if groups is not None:
            _groups = set(groups) if not isinstance(groups, str) else {groups}

        OpenhabItem = HABApp.openhab.items.OpenhabItem
        if _tags or _groups or metadata or metadata_value:
            if type is None:
                type = OpenhabItem
            if not issubclass(type, OpenhabItem):
                raise ValueError('Searching for tags, groups and metadata only works for OpenhabItem or its Subclasses')

        ret = []
        for item in HABApp.core.Items.get_all_items():  # type: HABApp.core.items.base_valueitem.BaseItem
            if type is not None and not isinstance(item, type):
                continue

            if name is not None and not name.search(item.name):
                continue

            if _tags is not None and not _tags.issubset(item.tags):
                continue

            if _groups is not None and not _groups.issubset(item.groups):
                continue

            if metadata is not None and not any(map(metadata.search, item.metadata)):
                continue

            if metadata_value is not None and not any(
                    map(metadata_value.search, map(lambda x: x[0], item.metadata.values()))):
                continue

            ret.append(item)
        return ret

    # -----------------------------------------------------------------------------------------------------------------
    # deprecated functions
    # -----------------------------------------------------------------------------------------------------------------
    def run_every(self, time, interval: Union[int, datetime.timedelta], callback, *args, **kwargs):
        warnings.warn('self.run_every is deprecated. Please use self.run.every', DeprecationWarning)
        return self.run.every(time, interval, callback, *args, **kwargs)

    def run_on_sun(self, sun_event: str, callback, *args, **kwargs):
        warnings.warn('self.run_on_sun is deprecated. Please use self.run.on_sunrise, self.run.on_sunset, ...',
                      DeprecationWarning)
        func = {'sunset': self.run.on_sunset, 'sunrise': self.run.on_sunrise,
                'dusk': self.run.on_sun_dusk, 'dawn': self.run.on_sun_dawn}
        return func[sun_event](callback, *args, **kwargs)

    def run_on_day_of_week(self, time: datetime.time, weekdays, callback, *args, **kwargs):
        warnings.warn('self.run_on_day_of_week is deprecated. Please use self.run.on_day_of_week', DeprecationWarning)
        return self.run.on_day_of_week(time, weekdays, callback, *args, **kwargs)

    def run_on_every_day(self, time: datetime.time, callback, *args, **kwargs):
        warnings.warn('self.run_on_every_day is deprecated. Please use self.run.on_every_day', DeprecationWarning)
        return self.run.on_every_day(time, callback, *args, **kwargs)

    def run_on_workdays(self, time: datetime.time, callback, *args, **kwargs):
        warnings.warn('self.run_on_workdays is deprecated. Please use self.run.on_workdays', DeprecationWarning)
        return self.run.on_workdays(time, callback, *args, **kwargs)

    def run_on_weekends(self, time: datetime.time, callback, *args, **kwargs):
        warnings.warn('self.run_on_weekends is deprecated. Please use self.run.on_weekends', DeprecationWarning)
        return self.run.on_weekends(time, callback, *args, **kwargs)

    def run_daily(self, callback, *args, **kwargs):
        warnings.warn('self.run_hourly is deprecated. Please use self.run.every', DeprecationWarning)
        start = datetime.timedelta(seconds=random.randint(0, 24 * 3600 - 1))
        return self.run.every(start, datetime.timedelta(days=1), callback, *args, **kwargs)

    def run_hourly(self, callback, *args, **kwargs) :
        warnings.warn('self.run_hourly is deprecated. Please use self.run.every_hour', DeprecationWarning)
        return self.run.every_hour(callback, *args, **kwargs)

    def run_minutely(self, callback, *args, **kwargs):
        warnings.warn('self.run_minutely is deprecated. Please use self.run.every_minute', DeprecationWarning)
        return self.run.every_minute(callback, *args, **kwargs)

    def run_at(self, date_time, callback, *args, **kwargs):
        warnings.warn('self.run_at is deprecated. Please use self.run.at', DeprecationWarning)
        return self.run.at(date_time, callback, *args, **kwargs)

    def run_in(self, seconds: Union[int, datetime.timedelta], callback, *args, **kwargs):
        warnings.warn('self.run_in is deprecated. Please use self.run.at', DeprecationWarning)
        return self.run.at(seconds, callback, *args, **kwargs)

    def run_soon(self, callback, *args, **kwargs):
        warnings.warn('self.run_soon is deprecated. Please use self.run.soon', DeprecationWarning)
        return self.run.soon(callback, *args, **kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    # internal functions
    # -----------------------------------------------------------------------------------------------------------------
    def _get_cb_name(self, callback):
        return f'{self.rule_name}.{callback.__name__}' if self.rule_name else None

    def _add_event_listener(self, listener: HABApp.core.EventBusListener) -> HABApp.core.EventBusListener:
        self.__event_listener.append(listener)
        HABApp.core.EventBus.add_listener(listener)
        return listener

    @HABApp.core.wrapper.log_exception
    def _check_rule(self):
        # We need items if we want to run the test
        if HABApp.core.Items.get_all_items():

            # Check if we have a valid item for all listeners
            for listener in self.__event_listener:

                # Internal topics - don't warn there
                if listener.topic in HABApp.core.const.topics.ALL:
                    continue

                # check if specific item exists
                if not HABApp.core.Items.item_exists(listener.topic):
                    log.warning(f'Item "{listener.topic}" does not exist (yet)! '
                                f'self.listen_event in "{self.rule_name}" may not work as intended.')

        # enable the scheduler
        self.run._scheduler.resume()


    @HABApp.core.wrapper.log_exception
    def _unload(self):

        # unload all functions
        for func in self.__unload_functions:
            try:
                func()
            except Exception as e:

                # try getting function name
                try:
                    name = f' in "{func.__name__}"'
                except AttributeError:
                    name = ''

                log.error(f'Error{name} while unloading "{self.rule_name}": {e}')

                # log traceback
                lines = traceback.format_exc().splitlines()
                del lines[1:3]  # see implementation in wrappedfunction.py why we do this
                for line in lines:
                    log.error(line)


@HABApp.core.wrapper.log_exception
def get_parent_rule() -> Rule:
    depth = 1
    while True:
        try:
            frm = sys._getframe(depth)
        except ValueError:
            raise RuntimeError('Could not find parent rule!') from None

        __vars = frm.f_locals
        depth += 1
        if 'self' in __vars:
            rule = __vars['self']
            if isinstance(rule, Rule):
                return rule



import datetime
import logging
from typing import Any, Dict, FrozenSet, Optional

from immutables import Map

import HABApp
from HABApp.core.wrapper import process_exception
from HABApp.openhab.definitions.values import QuantityValue, RawValue
from HABApp.openhab.items import ColorItem, ContactItem, DatetimeItem, DimmerItem, GroupItem, ImageItem, LocationItem, \
    NumberItem, PlayerItem, RollershutterItem, StringItem, SwitchItem
from HABApp.openhab.items.base_item import MetaData

log = logging.getLogger('HABApp.openhab')


def map_item(name: str, type: str, value: Optional[str],
             tags: FrozenSet[str], groups: FrozenSet[str], metadata: Optional[Dict[str, Dict[str, Any]]]) -> \
        Optional['HABApp.openhab.items.OpenhabItem']:
    try:
        assert isinstance(type, str)
        assert value is None or isinstance(value, str)

        if value == 'NULL' or value == 'UNDEF':
            value = None

        # map Metadata
        if metadata is not None:
            meta = Map({k: MetaData(v['value'], Map(v.get('config', {}))) for k, v in metadata.items()})
        else:
            meta = Map()

        # Quantity types are like this: Number:Temperature and have a unit set: "12.3 °C".
        # We have to remove the dimension from the type and remove the unit from the value
        if ':' in type:
            type, dimension = type.split(':')
            # if the item is not initialized its None and has no dimension
            if value is not None:
                value, _ = QuantityValue.split_unit(value)

        # Specific classes
        if type == "Switch":
            return SwitchItem(name, value, tags=tags, groups=groups, metadata=meta)

        if type == "String":
            return StringItem(name, value, tags=tags, groups=groups, metadata=meta)

        if type == "Contact":
            return ContactItem(name, value, tags=tags, groups=groups, metadata=meta)

        if type == "Rollershutter":
            if value is None:
                return RollershutterItem(name, value, tags=tags, groups=groups, metadata=meta)
            return RollershutterItem(name, float(value), tags=tags, groups=groups, metadata=meta)

        if type == "Dimmer":
            if value is None:
                return DimmerItem(name, value, tags=tags, groups=groups, metadata=meta)
            return DimmerItem(name, float(value), tags=tags, groups=groups, metadata=meta)

        if type == "Number":
            if value is None:
                return NumberItem(name, value, tags=tags, groups=groups, metadata=meta)

            # Number items can be int or float
            try:
                return NumberItem(name, int(value), tags=tags, groups=groups, metadata=meta)
            except ValueError:
                return NumberItem(name, float(value), tags=tags, groups=groups, metadata=meta)

        if type == "DateTime":
            if value is None:
                return DatetimeItem(name, value, tags=tags, groups=groups, metadata=meta)
            # Todo: remove this once we go >= OH3.1
            # Previous OH versions used a datetime string like this:
            # 2018-11-19T09:47:38.284+0100
            # OH 3.1 uses
            # 2021-04-10T22:00:43.043996+0200
            if len(value) == 28:
                value = value.replace('+', '000+')
            dt = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f%z')
            # all datetimes from openhab have a timezone set so we can't easily compare them
            # --> TypeError: can't compare offset-naive and offset-aware datetimes
            dt = dt.astimezone(tz=None)   # Changes datetime object so it uses system timezone
            dt = dt.replace(tzinfo=None)  # Removes timezone awareness
            return DatetimeItem(name, dt, tags=tags, groups=groups, metadata=meta)

        if type == "Color":
            if value is None:
                return ColorItem(name, tags=tags, groups=groups, metadata=meta)
            return ColorItem(name, *(float(k) for k in value.split(',')), tags=tags, groups=groups, metadata=meta)

        if type == "Image":
            img = ImageItem(name, tags=tags, groups=groups, metadata=meta)
            if value is None:
                return img
            img.set_value(RawValue(value))
            return img

        if type == "Group":
            return GroupItem(name, value, tags=tags, groups=groups, metadata=meta)

        if type == "Location":
            return LocationItem(name, value, tags=tags, groups=groups, metadata=meta)

        if type == "Player":
            return PlayerItem(name, value, tags=tags, groups=groups, metadata=meta)

        raise ValueError(f'Unknown Openhab type: {type} for {name}')

    except Exception as e:
        process_exception('map_items', e, logger=log)
        return None



import time
from enum import Enum


class _MissingType(Enum):
    MISSING = object()

    def __str__(self):
        return '<Missing>'


# todo: add type final if we go >= python 3.8
MISSING = _MissingType.MISSING
STARTUP = time.time()



from asyncio import create_task
from typing import Union

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.core import EventBus, Items
from HABApp.core.Items import ItemNotFoundException
from HABApp.core.events import ValueUpdateEvent
from HABApp.core.logger import log_warning
from HABApp.core.wrapper import process_exception
from HABApp.openhab.connection_handler import http_connection
from HABApp.openhab.events import GroupItemStateChangedEvent, ItemAddedEvent, ItemRemovedEvent, ItemUpdatedEvent, \
    ThingStatusInfoEvent
from HABApp.openhab.item_to_reg import add_to_registry, remove_from_registry
from HABApp.openhab.map_events import get_event
from HABApp.openhab.map_items import map_item

log = http_connection.log


def on_sse_event(event_dict: dict):
    try:
        # Lookup corresponding OpenHAB event
        event = get_event(event_dict)

        # Update item in registry BEFORE posting to the event bus
        # so the items have the correct state when we process the event in a rule
        try:
            if isinstance(event, ValueUpdateEvent):
                __item = Items.get_item(event.name)  # type: HABApp.core.items.base_item.BaseValueItem
                __item.set_value(event.value)
                EventBus.post_event(event.name, event)
                return None

            if isinstance(event, ThingStatusInfoEvent):
                __thing = Items.get_item(event.name)   # type: HABApp.openhab.items.Thing
                __thing.process_event(event)
                EventBus.post_event(event.name, event)
                return None

            # Workaround because there is no GroupItemStateEvent
            if isinstance(event, GroupItemStateChangedEvent):
                __item = Items.get_item(event.name)  # type: HABApp.openhab.items.GroupItem
                __item.set_value(event.value)
                EventBus.post_event(event.name, event)
                return None
        except ItemNotFoundException:
            log_warning(log, f'Received {event.__class__.__name__} for {event.name} but item does not exist!')

            # Post the event anyway
            EventBus.post_event(event.name, event)
            return None

        if isinstance(event, ItemRemovedEvent):
            remove_from_registry(event.name)
            EventBus.post_event(event.name, event)
            return None

        # These events require that we query openhab because of the metadata so we have to do it in a task
        # They also change the item registry
        if isinstance(event, (ItemAddedEvent, ItemUpdatedEvent)):
            create_task(item_event(event))
            return None

        HABApp.core.EventBus.post_event(event.name, event)
    except Exception as e:
        process_exception(func=on_sse_event, e=e)
        return None


async def item_event(event: Union[ItemAddedEvent, ItemUpdatedEvent]):
    name = event.name

    # Since metadata is not part of the event we have to request it
    cfg = await HABApp.openhab.interface_async.async_get_item(name, metadata='.+')

    new_item = map_item(name, event.type, None, event.tags, event.groups, metadata=cfg.get('metadata'))
    if new_item is None:
        return None

    add_to_registry(new_item)
    # Send Event to Event Bus
    HABApp.core.EventBus.post_event(name, event)
    return None



from pathlib import Path
from typing import List, Iterable, TYPE_CHECKING

if TYPE_CHECKING:
    import HABApp


def list_files(folder: Path, file_filter: 'HABApp.core.files.watcher.file_watcher.EventFilterBase',
               recursive: bool = False) -> List[Path]:
    # glob is much quicker than iter_dir()
    files = folder.glob('**/*' if recursive else '*')
    return sorted(filter(lambda x: file_filter.notify(str(x)), files), key=lambda x: x.relative_to(folder))


def sort_files(files: Iterable[Path]) -> List[Path]:
    return sorted(files)



from typing import Any, Callable, List, Optional, Union

from HABApp.core.event_bus_listener import EventBusListener
from HABApp.core.events import AllEvents, EventFilter
from HABApp.core.items.base_valueitem import BaseItem


class EventListenerCreator:
    def __init__(self, item: BaseItem, callback: Callable[[Any], Any],  event_filter: EventFilter):
        self.item = item
        self.callback = callback
        self.event_filter = event_filter

    def listen(self) -> EventBusListener:
        return self.item.listen_event(self.callback, self.event_filter)


class NoUpdateEventListenerCreator:
    def __init__(self, item: BaseItem, callback: Callable[[Any], Any], secs: Union[int, float]):
        self.item = item
        self.callback = callback
        self.secs = secs

    def listen(self) -> EventBusListener:
        return self.item.watch_update(self.secs).listen_event(self.callback)


class NoChangeEventListenerCreator:
    def __init__(self, item: BaseItem, callback: Callable[[Any], Any], secs: Union[int, float]):
        self.item = item
        self.callback = callback
        self.secs = secs

    def listen(self) -> EventBusListener:
        return self.item.watch_change(self.secs).listen_event(self.callback)


class EventListenerGroup:
    """Helper to create/cancel multiple event listeners simultaneously
    """
    def __init__(self, default_callback: Optional[Callable[[Any], Any]] = None, default_event_filter=AllEvents,
                 default_seconds: Optional[Union[int, float]] = None):
        self._items: List[Union[EventListenerCreator, NoUpdateEventListenerCreator, NoChangeEventListenerCreator]] = []
        self._subs: List[EventBusListener] = []

        self._is_active = False

        self._default_callback = default_callback
        self._default_event_filter = default_event_filter
        self._default_seconds = default_seconds

    @property
    def active(self) -> bool:
        """

        :return: True if the listeners are currently active
        """
        return self._is_active

    def listen(self):
        """Create all event listeners. If the event listeners are already active this will do nothing.
        """
        if self._is_active:
            return None
        self._is_active = True

        for obj in self._items:
            self._subs.append(obj.listen())

    def cancel(self):
        """Cancel the active event listeners. If the event listeners are not active this will do nothing.
        """
        if not self._is_active:
            return None
        self._is_active = False

        while self._subs:
            self._subs.pop().cancel()

    def add_listener(self, item: BaseItem, callback: Optional[Callable[[Any], Any]] = None,
                     event_filter: Optional[EventFilter] = None) -> 'EventListenerGroup':
        """Add an event listener to the group

        :param item: Item
        :param callback: Callback or default callback if omitted
        :param event_filter: Event filter of default event filter if omitted
        :return: self
        """
        if callback is None:
            callback = self._default_callback
        if callback is None:
            raise ValueError('No callback passed and no default callback specified in __init__')

        if event_filter is None:
            event_filter = self._default_event_filter

        obj = EventListenerCreator(item, callback, event_filter if event_filter is not None else AllEvents)
        self._items.append(obj)

        if self._is_active:
            self._subs.append(obj.listen())
        return self

    def add_no_update_watcher(self, item: BaseItem, callback: Optional[Callable[[Any], Any]] = None,
                              seconds: Optional[Union[int, float]] = None) -> 'EventListenerGroup':
        """Add an no update watcher to the group. On ``listen`` this will create a no update watcher and
         the corresponding event listener that will trigger the callback

        :param item: Item
        :param callback: Callback or default callback if omitted
        :param seconds: No update time for the no update watcher or default seconds if omitted
        :return: self
        """
        if callback is None:
            callback = self._default_callback
        if seconds is None:
            seconds = self._default_seconds

        if callback is None:
            raise ValueError('No callback passed and no default callback specified in __init__')
        if seconds is None:
            raise ValueError('No seconds passed and no default seconds specified in __init__')

        obj = NoUpdateEventListenerCreator(item, callback, seconds)
        self._items.append(obj)

        if self._is_active:
            self._subs.append(obj.listen())
        return self

    def add_no_change_watcher(self, item: BaseItem, callback: Optional[Callable[[Any], Any]] = None,
                              seconds: Optional[Union[int, float]] = None) -> 'EventListenerGroup':
        """Add an no change watcher to the group. On ``listen`` this this will create a no change watcher and
         the corresponding event listener that will trigger the callback

        :param item: Item
        :param callback: Callback or default callback if omitted
        :param seconds: No update time for the no change watcher or default seconds if omitted
        :return: self
        """
        if callback is None:
            callback = self._default_callback
        if seconds is None:
            seconds = self._default_seconds

        if callback is None:
            raise ValueError('No callback passed and no default callback specified in __init__')
        if seconds is None:
            raise ValueError('No seconds passed and no default seconds specified in __init__')

        obj = NoChangeEventListenerCreator(item, callback, seconds)
        self._items.append(obj)

        if self._is_active:
            self._subs.append(obj.listen())
        return self



import logging.config

import eascheduler
import voluptuous
from EasyCo import ConfigContainer, ConfigEntry

log = logging.getLogger('HABApp.Config')


class Location(ConfigContainer):
    latitude: float = ConfigEntry(default=0.0, validator=voluptuous.Any(float, int))
    longitude: float = ConfigEntry(default=0.0, validator=voluptuous.Any(float, int))
    elevation: float = ConfigEntry(default=0.0, validator=voluptuous.Any(float, int))

    def __init__(self):
        super().__init__()

    def on_all_values_set(self):
        log.debug(f'Local Timezone: {eascheduler.const.local_tz}')
        eascheduler.set_location(self.latitude, self.longitude, self.elevation)



import logging
from typing import Union

import HABApp
from HABApp.core.const.topics import FILES as T_FILES
from HABApp.core.events.habapp_events import RequestFileUnloadEvent, RequestFileLoadEvent

log = logging.getLogger('HABApp.Files')


async def _process_event(event: Union[RequestFileUnloadEvent, RequestFileLoadEvent]):
    name = event.name
    await HABApp.core.files.manager.process_file(name, HABApp.core.files.folders.get_path(name))


async def setup_file_manager():
    # Setup events so we can process load/unload
    HABApp.core.EventBus.add_listener(
        HABApp.core.EventBusListener(T_FILES, HABApp.core.WrappedFunction(_process_event), RequestFileUnloadEvent)
    )
    HABApp.core.EventBus.add_listener(
        HABApp.core.EventBusListener(T_FILES, HABApp.core.WrappedFunction(_process_event), RequestFileLoadEvent)
    )



import logging

import HABApp
from HABApp.core import Items
from HABApp.core.wrapper import ignore_exception
from HABApp.openhab.map_items import map_item
from ._plugin import OnConnectPlugin
from ..interface_async import async_get_items, async_get_things
from HABApp.openhab.item_to_reg import add_to_registry, fresh_item_sync

log = logging.getLogger('HABApp.openhab.items')


class LoadAllOpenhabItems(OnConnectPlugin):

    @ignore_exception
    async def on_connect_function(self):
        data = await async_get_items(disconnect_on_error=True, all_metadata=True)
        if data is None:
            return None

        fresh_item_sync()

        found_items = len(data)
        for _dict in data:
            item_name = _dict['name']
            new_item = map_item(item_name, _dict['type'], _dict['state'],
                                frozenset(_dict['tags']), frozenset(_dict['groupNames']),
                                _dict.get('metadata', {}))   # type: HABApp.openhab.items.OpenhabItem
            if new_item is None:
                continue
            add_to_registry(new_item, True)

        # remove items which are no longer available
        ist = set(Items.get_all_item_names())
        soll = {k['name'] for k in data}
        for k in ist - soll:
            if isinstance(Items.get_item(k), HABApp.openhab.items.OpenhabItem):
                Items.pop_item(k)

        log.info(f'Updated {found_items:d} Items')

        # try to update things, too
        data = await async_get_things(disconnect_on_error=True)
        if data is None:
            return None

        Thing = HABApp.openhab.items.Thing
        for t_dict in data:
            name = t_dict['UID']
            try:
                thing = Items.get_item(name)
                if not isinstance(thing, Thing):
                    log.warning(f'Item {name} has the wrong type ({type(thing)}), expected Thing')
                    thing = Thing(name)
            except Items.ItemNotFoundException:
                thing = Thing(name)

            thing.status = t_dict['statusInfo']['status']
            Items.add_item(thing)

        # remove things which were deleted
        ist = set(Items.get_all_item_names())
        soll = {k['UID'] for k in data}
        for k in ist - soll:
            if isinstance(Items.get_item(k), Thing):
                Items.pop_item(k)

        log.info(f'Updated {len(data):d} Things')
        return None


PLUGIN_LOAD_ITEMS = LoadAllOpenhabItems.create_plugin()



from typing import Any, Dict

from pydantic import Field

from .base import RestBase


class ItemChannelLinkDefinition(RestBase):
    item_name: str = Field(alias='itemName')
    channel_uid: str = Field(alias='channelUID')
    configuration: Dict[str, Any] = {}

    # This field is OH3 only
    # Todo: Remove this comment once we go OH3
    editable: bool = False


class LinkNotFoundError(Exception):
    pass



import logging
from typing import Dict, Set, Tuple

import HABApp
from HABApp.core import Items
from HABApp.core.logger import log_warning

log = logging.getLogger('HABApp.openhab.items')


def add_to_registry(item: 'HABApp.openhab.items.OpenhabItem', set_value=False):
    name = item.name
    for grp in item.groups:
        MEMBERS.setdefault(grp, set()).add(name)

    if not Items.item_exists(name):
        return Items.add_item(item)

    existing = Items.get_item(name)
    if isinstance(existing, item.__class__):
        # If we load directly through the API and not through an event we have to set the value
        if set_value:
            existing.set_value(item.value)

        # remove old groups
        for grp in set(existing.groups) - set(item.groups):
            MEMBERS.get(grp, set()).discard(name)

        # same type - it was only an item update (e.g. label)!
        existing.tags = item.tags
        existing.groups = item.groups
        return None

    log_warning(log, f'Item type changed from {existing.__class__} to {item.__class__}')

    # Replace existing item with the updated definition
    Items.pop_item(name)
    Items.add_item(item)


def remove_from_registry(name: str):
    if not Items.item_exists(name):
        return None

    item = Items.get_item(name)  # type: HABApp.openhab.items.OpenhabItem
    for grp in item.groups:
        MEMBERS.get(grp, set()).discard(name)

    if isinstance(item, HABApp.openhab.items.GroupItem):
        MEMBERS.pop(name, None)

    Items.pop_item(name)
    return None


MEMBERS: Dict[str, Set[str]] = {}


def fresh_item_sync():
    MEMBERS.clear()


def get_members(group_name: str) -> Tuple['HABApp.openhab.items.OpenhabItem', ...]:
    ret = []
    for name in MEMBERS.get(group_name, []):
        ret.append(Items.get_item(name))
    return tuple(sorted(ret))



from contextvars import ContextVar as _ContextVar
from typing import Callable as _Callable


async_context = _ContextVar('async_ctx')


class AsyncContextError(Exception):
    def __init__(self, func: _Callable) -> None:
        super().__init__()
        self.func: _Callable = func

    def __str__(self):
        return f'Function "{self.func.__name__}" may not be called from an async context!'



import logging
import sys
from pathlib import Path

from EasyCo import ConfigEntry, ConfigFile, PathContainer

from ._conf_location import Location
from ._conf_mqtt import Mqtt
from ._conf_openhab import Openhab
from .platform_defaults import get_log_folder

log = logging.getLogger('HABApp.Config')


class Directories(PathContainer):
    logging: Path = ConfigEntry(get_log_folder(Path('log')), description='Folder where the logs will be written to')
    rules: Path = ConfigEntry(Path('rules'), description='Folder from which the rule files will be loaded')
    param: Path = ConfigEntry(Path('params'), description='Folder from which the parameter files will be loaded')
    config: Path = ConfigEntry(Path('config'), description='Folder from which configuration files '
                                                           '(e.g. for textual thing configuration) will be loaded')
    lib: Path = ConfigEntry(Path('lib'), description='Folder where additional libraries can be placed')

    def on_all_values_set(self):

        # Configuration folder of HABApp can not be one of the configured folders
        for name, path in {attr: getattr(self, attr) for attr in ('rules', 'param', 'config')}.items():
            if path == self.parent_folder:
                msg = f'Path for {name} can not be the same as the path for the HABApp config! ({path})'
                log.error(msg)
                sys.exit(msg)

        try:
            # create folder structure if it does not exist
            if not self.rules.is_dir():
                self.rules.mkdir()
            if not self.logging.is_dir():
                self.logging.mkdir()
            if not self.config.is_dir():
                log.info(f'Manual thing configuration disabled! Folder {self.config} does not exist!')

            # add path for libraries
            if self.lib.is_dir():
                lib_path = str(self.lib)
                if lib_path not in sys.path:
                    sys.path.insert(0, lib_path)
                    log.debug(f'Added library folder "{lib_path}" to system path')
        except Exception as e:
            log.error(e)
            print(e)


class HABAppConfig(ConfigFile):
    location = Location()
    directories = Directories()

    mqtt = Mqtt()
    openhab = Openhab()


CONFIG: HABAppConfig = HABAppConfig()



from __future__ import annotations

from enum import Enum, auto


class FileState(Enum):
    LOADED = auto()
    FAILED = auto()

    DEPENDENCIES_OK = auto()
    DEPENDENCIES_MISSING = auto()
    PROPERTIES_ERROR = auto()

    PENDING = auto()
    REMOVED = auto()

    def __str__(self):
        return str(self.name)



from __future__ import annotations

from pathlib import Path
from typing import Type, Dict

from HABApp.core.files.file import HABAppFile
from HABApp.core.files.file.properties import get_properties

FILE_TYPES: Dict[str, Type[HABAppFile]] = {}


def register_file_type(prefix: str, cls: Type[HABAppFile]):
    assert prefix not in FILE_TYPES

    assert cls.LOGGER
    assert cls.LOAD_FUNC
    assert cls.UNLOAD_FUNC

    FILE_TYPES[prefix] = cls


def create_file(name: str, path: Path) -> HABAppFile:
    for prefix, cls in FILE_TYPES.items():
        if name.startswith(prefix):
            break
    else:
        raise ValueError(f'Unknown file type for "{name}"!')

    with path.open('r', encoding='utf-8') as f:
        txt = f.read(10 * 1024)
    return cls(name, path, get_properties(txt))



import datetime
from typing import Any, FrozenSet, Mapping, NamedTuple, Optional

from immutables import Map

from HABApp.core.const import MISSING
from HABApp.core.items.base_valueitem import BaseValueItem
from HABApp.openhab.interface import get_persistence_data, post_update, send_command


class MetaData(NamedTuple):
    value: str
    config: Mapping[str, Any] = Map()


class OpenhabItem(BaseValueItem):
    """Base class for items which exists in OpenHAB.
    """

    def __init__(self, name: str, initial_value=None,
                 tags: FrozenSet[str] = frozenset(), groups: FrozenSet[str] = frozenset(),
                 metadata: Mapping[str, MetaData] = Map()):
        super().__init__(name, initial_value)
        self.tags: FrozenSet[str] = tags
        self.groups: FrozenSet[str] = groups
        self.metadata: Mapping[str, MetaData] = metadata

    def oh_send_command(self, value: Any = MISSING):
        """Send a command to the openHAB item

        :param value: (optional) value to be sent. If not specified the item value will be used.
        """
        send_command(self.name, self.value if value is MISSING else value)

    def oh_post_update(self, value: Any = MISSING):
        """Post an update to the openHAB item

        :param value: (optional) value to be posted. If not specified the item value will be used.
        """
        post_update(self.name, self.value if value is MISSING else value)

    def get_persistence_data(self, persistence: Optional[str] = None,
                             start_time: Optional[datetime.datetime] = None,
                             end_time: Optional[datetime.datetime] = None):
        """Query historical data from the OpenHAB persistence service

        :param persistence: name of the persistence service (e.g. ``rrd4j``, ``mapdb``). If not set default will be used
        :param start_time: return only items which are newer than this
        :param end_time: return only items which are older than this
        """

        return get_persistence_data(
            self.name, persistence, start_time, end_time
        )
